import {Mutex} from './mutex';

/** A condition variable.
 *
 * A condition variable allows multiple coroutines to wait for an event to be
 * triggered.
 */
export class Condvar {
    private _mu: Mutex;
    private _waiters: Array<() => void>;

    /** Creates a new condition variable.
     *
     * Condition variables must be associated with mutexes, which protects the
     * condition variable itself.
     *
     * @param mu a mutex to use
     */
    public constructor(mu: Mutex) {
        this._mu = mu;
        this._waiters = [];
    }

    /** Waits on the condition variable to be notified.
     *
     * This will unlock the underlying mutex and, upon being awoken by another
     * thread, lock the mutex again.
     *
     * Spurious wake-ups are not possible, so it is safe to assume that if a
     * coroutine was awoken, another coroutine must have notified it.
     *
     * @returns a promise to wait on
     */
    public wait(): Promise<void> {
        return new Promise<void>((resolve, reject) => {
            try {
                this._mu.unlock();
                this._waiters.push(resolve);
            } catch (e) {
                reject(e);
            }
        });
    }

    /** Notifies one coroutine to wake up.
     *
     * It will run one waiting continuation.
     *
     * This operation does not need to be waited on.
     */
    public notify(): void {
        if (this._waiters.length > 0) {
            let waiter = this._waiters.shift();
            this._mu.lock().then(() => {
                waiter();
            });
        }
    }

    /** Notifies all coroutines to wake up.
     *
     * It will run all waiting continuations.
     *
     * This operation does not need to be waited on.
     */
    public notifyAll(): void {
        while (this._waiters.length > 0) {
            this.notify();
        }
    }
}
