/** A mutual exclusion device.
 *
 * A mutex allows only one coroutine to be executing from the time of its
 * locking to the time of its unlocking.
 */
export class Mutex {
    private _locked: boolean;
    private _waiters: Array<() => void>;

    /** Creates a mutex.
     */
    public constructor() {
        this._locked = false;
        this._waiters = [];
    }

    /** Tries to lock the mutex.
     *
     * This operation does not need to be waited on.
     *
     * @returns whether or not the operation succeeded
     */
    public tryLock(): boolean {
        if (this._locked) {
            return false;
        }
        this._locked = true;
        return true;
    }

    /** Locks the mutex.
     *
     * The returned promise will block until the mutex is not locked.
     *
     * @returns a promise to wait on
     */
    public lock(): Promise<void> {
        return new Promise<void>((resolve, reject) => {
            if (this.tryLock()) {
                resolve();
            } else {
                this._waiters.push(resolve);
            }
        });
    }

    /** Unlocks the mutex.
     *
     * This may cause the execution of the first continuation waiting on the
     * lock, if there is any.
     *
     * This operation does not need to be waited on.
     */
    public unlock(): void {
        if (!this._locked) {
            throw new Error('attempted unlock of unlocked mutex');
        }

        if (this._waiters.length > 0) {
            this._waiters.shift()();
        } else {
            this._locked = false;
        }
    }
}
