enum LockMode {
    NONE,
    SHARED,
    EXCLUSIVE
}

/** A readers–writer lock.
 *
 * A readers–writer lock allows concurrent access for read-only operations,
 * but exclusive access for write operations, i.e. multiple coroutines can read
 * in parallel, but only one coroutine can write at a time.
 *
 * If a writer is writing data, all writers and readers will not be allowed to
 * progress until the lock is released. If readers are reading data, no writer
 * will be allowed to progress until there are no more readers reading.
 *
 * This implementation does not support lock upgrading.
 */
export class RwLock {
    private _lockMode: LockMode;
    private _readLocks: number;
    private _readWaiters: Array<() => void>;
    private _writeWaiters: Array<() => void>;

    /** Creates a readers–writer lock.
     */
    public constructor() {
        this._lockMode = LockMode.NONE;
        this._readLocks = 0;
        this._readWaiters = [];
        this._writeWaiters = [];
    }

    /** Tries to acquire a reader lock.
     *
     * This operation does not need to be waited on.
     *
     * @returns whether or not the operation succeeded
     */
    public tryReadLock(): boolean {
        if (this._lockMode === LockMode.EXCLUSIVE) {
            return false;
        }
        this._lockMode = LockMode.SHARED;
        ++this._readLocks;
        return true;
    }

    /** Acquires a read lock.
     *
     * The returned promise will block until no writers are active.
     *
     * @returns a promise to wait on
     */
    public readLock(): Promise<void> {
        return new Promise<void>((resolve, reject) => {
            if (this.tryReadLock()) {
                resolve();
            } else {
                this._readWaiters.push(resolve);
            }
        });
    }

    /** Tries to acquire the writer lock.
     *
     * This operation does not need to be waited on.
     *
     * @returns whether or not the operation succeeded
     */
    public tryWriteLock(): boolean {
        if (this._lockMode !== LockMode.NONE) {
            return false;
        }
        this._lockMode = LockMode.EXCLUSIVE;
        return true;
    }

    /** Acquires a write lock.
     *
     * The returned promise will block until no readers or writers are active.
     *
     * @returns a promise to wait on
     */
    public writeLock(): Promise<void> {
       return new Promise<void>((resolve, reject) => {
            if (this.tryWriteLock()) {
                resolve();
            } else {
                this._writeWaiters.push(resolve);
            }
        });
    }

    /** Unlocks the readers–writer lock.
     *
     * This may cause the execution of continuations waiting on the lock and
     * re-locking:
     *
     * * If all readers had become inactive, the first waiting writer will be
     *   resumed and the writer lock will be acquired.
     *
     * * If the writer had become inactive, all readers will be resumed and
     *   reader locks acquried, if any are waiting. If there are no readers
     *   waiting but a writer is waiting, the writer will be resumed and the
     *   writer lock acquired.
     *
     * If none of the above conditions are met, the readers–writer lock will be
     * unlocked.
     *
     * This operation does not need to be waited on.
     */
    public unlock(): void {
        switch (this._lockMode) {
            case LockMode.NONE:
                throw new Error('attempted unlock of unlocked rwlock');

            case LockMode.SHARED:
                --this._readLocks;
                if (this._readLocks === 0) {
                    if (this._writeWaiters.length > 0) {
                        // Run one writer.
                        this._lockMode = LockMode.EXCLUSIVE;
                        this._writeWaiters.shift()();
                    } else {
                        // Fully unlocked.
                        this._lockMode = LockMode.NONE;
                    }
                }
                break;
            case LockMode.EXCLUSIVE:
                if (this._readWaiters.length > 0) {
                    // Run all readers.
                    this._lockMode = LockMode.SHARED;
                    this._readLocks = this._readWaiters.length;

                    let waiters = this._readWaiters.slice();
                    this._readWaiters.length = 0;

                    for (let waiter of waiters) {
                        waiter();
                    }
                } else if (this._writeWaiters.length > 0) {
                    // Run one writer.
                    this._writeWaiters.shift()();
                } else {
                    // Fully unlocked.
                    this._lockMode = LockMode.NONE;
                }
                break;
        }
    }
}
