/** A counting semaphore.
 *
 * A counting semaphore controls access to a resource by multiple coroutines.
 * It contains an internal counter which can be incremented and decremented.
 * If, at any time, the counter is decremented below zero, a coroutine must wait
 * until other coroutines increments it above zero.
 */
export class Semaphore {
    private _value: number;
    private _waiters: Array<() => void>;

    /** Creates a counting semaphore.
     *
     * @param n the initial value of the counting semaphore
     */
    public constructor(n: number) {
        this._value = n;
        this._waiters = [];
    }

    /** Tries to decrement the counting semaphore.
     *
     * This operation does not need to be waited on.
     *
     * @returns whether or not the operation succeeded
     */
    public tryWait(): boolean {
        if (this._value <= 0) {
            return false;
        }
        --this._value;
        return true;
    }

    /** Decrements the counting semaphore and waits.
     *
     * The returned promise will block until the counting semaphore has a value
     * greater than zero.
     *
     * @returns a promise to wait on
     */
    public wait(): Promise<void> {
        return new Promise<void>((resolve, reject) => {
            if (this.tryWait()) {
                resolve();
            } else {
                this._waiters.push(resolve);
            }
        });
    }

    /** Increments the counting semaphore.
     *
     * This may cause the execution of continuations waiting on the counting
     * semaphore, if there are any.
     *
     * This operation does not need to be waited on.
     */
    public signal(): void {
        if (this._waiters.length > 0 && this._value == 0) {
            this._waiters.shift()();
        } else {
            ++this._value;
        }
    }

    /** The current value of the counting semaphore.
     */
    public get value(): number {
        return this._value;
    }
}
