/// <reference path="../typings/index.d.ts" />

const assert: any = require('assert');

import {RwLock} from '../lib/rwlock';

describe('RwLock', function () {
    it('should be write lockable when unlocked', function () {
        let rwlock = new RwLock();
        assert(rwlock.tryWriteLock());
    });

    it('should be read lockable when unlocked', function () {
        let rwlock = new RwLock();
        assert(rwlock.tryReadLock());
    });

    it('should not be unlockable when unlocked', function () {
        let rwlock = new RwLock();
        assert.throws(() => rwlock.unlock());
    });

    it('should be read lockable multiple times when unlocked', function () {
        let rwlock = new RwLock();
        assert(rwlock.tryReadLock());
        assert(rwlock.tryReadLock());
    });

    it('should not be write lockable multiple times when unlocked', function () {
        let rwlock = new RwLock();
        assert(rwlock.tryWriteLock());
        assert(!rwlock.tryWriteLock());
    });

    it('should not be write lockable after read lock', function () {
        let rwlock = new RwLock();
        assert(rwlock.tryReadLock());
        assert(!rwlock.tryWriteLock());
    });

    it('should not be read lockable after write lock', function () {
        let rwlock = new RwLock();
        assert(rwlock.tryWriteLock());
        assert(!rwlock.tryReadLock());
    });

    it('should be write lockable when after write unlock', function () {
        let rwlock = new RwLock();
        assert(rwlock.tryWriteLock());
        rwlock.unlock();
        assert(rwlock.tryWriteLock());
    });

    it('should be write lockable when after read unlock', function () {
        let rwlock = new RwLock();
        assert(rwlock.tryReadLock());
        rwlock.unlock();
        assert(rwlock.tryWriteLock());
    });

    it('should be read lockable when after write unlock', function () {
        let rwlock = new RwLock();
        assert(rwlock.tryWriteLock());
        rwlock.unlock();
        assert(rwlock.tryReadLock());
    });

    it('should be read lockable when after read unlock', function () {
        let rwlock = new RwLock();
        assert(rwlock.tryReadLock());
        rwlock.unlock();
        assert(rwlock.tryReadLock());
    });

    it('should run all reader callbacks after write unlock', function (done) {
        let n = 3;

        function countdown() {
            --n;
            if (n === 0) {
                done();
            }
        }

        let rwlock = new RwLock();
        assert(rwlock.tryWriteLock());
        rwlock.readLock().then(() => countdown());
        rwlock.readLock().then(() => countdown());
        rwlock.readLock().then(() => countdown());
        rwlock.unlock();
    });

    it('should run one writer callback after write unlock', function (done) {
        let rwlock = new RwLock();
        assert(rwlock.tryWriteLock());
        rwlock.writeLock().then(() => done());
        rwlock.writeLock().then(() => assert.ok(false, 'ran wrong writer callback'));
        rwlock.unlock();
    });

    it('should run one writer callback after read unlock', function (done) {
        let rwlock = new RwLock();
        assert(rwlock.tryReadLock());
        rwlock.writeLock().then(() => done());
        rwlock.writeLock().then(() => assert.ok(false, 'ran wrong writer callback'));
        rwlock.unlock();
    });

    it('should run all reader callbacks', function (done) {
        let n = 3;

        function countdown() {
            --n;
            if (n === 0) {
                done();
            }
        }

        let rwlock = new RwLock();
        rwlock.readLock().then(() => countdown());
        rwlock.readLock().then(() => countdown());
        rwlock.readLock().then(() => countdown());
        rwlock.unlock();
    });
});
